@@ -1,13 +1,14 @@
-__version__ = "0.5.5"
+__version__ = "1.0.0"
 
 import collections
-import os
 import json
 import serial
 import struct
 import threading
 import time
 
+from . import parameters
+
 # status codes dict, input code, get string
 pp_status_codes = { 0: 'PP_STATUS_OK',
                     1: 'PP_STATUS_PROCESS_CLAIMED',
@@ -126,6 +127,7 @@ class instrument(object):
     comport (str): COM port on which the instrument is connected (e.g. 'COM1' or '/dev/ttyUSB0').
     address (int, optional): Address of the instrument, default = 128 for local instrument.
     baudrate (int, optional): Baudrate to use for communication.
+    channel (int, optional): Channel to use for communication.
     serial_class (obj, optional): Custom serial class to be used for serial communication with the instrument.
 
   Attributes:
@@ -135,9 +137,10 @@ class instrument(object):
     db (obj): Instance of the propar database, for conversion from DDE number to process, parameter number.
   """
 
-  def __init__(self, comport, address=0x80, baudrate=38400, serial_class=serial.Serial):
+  def __init__(self, comport, address=0x80, baudrate=38400, channel=1, serial_class=serial.Serial):
     self.address = address
     self.comport = comport
+    self.channel = channel
     if comport in _PROPAR_MASTERS:    
       # Master already created previously
       self.master = _PROPAR_MASTERS[comport]
@@ -147,11 +150,30 @@ class instrument(object):
       _PROPAR_MASTERS[comport] = self.master
     self.db = self.master.db
 
-  def readParameter(self, dde_nr):
+  def __modify_parameter_channel(self, parm, channel=None):
+    """Adjust the parameter definition for current channel.
+    
+    Args:
+      parm (dict): Parameter to modify.
+      channel (int, optional): Channel to use for communication (if none, self.channel is used).
+
+    Returns:
+      Modified parameter.
+    """
+    if channel == None:
+      channel = self.channel
+    if channel >= 1 and channel <= 16:
+      if parm['proc_nr'] in [1, 33, 65, 97, 104]:
+        parm = dict(parm)
+        parm['proc_nr'] += channel - 1
+    return parm
+
+  def readParameter(self, dde_nr, channel=None):
     """Read a single parameter indicated by DDE nr.
 
     Args:
       dde_nr (int): FlowDDE parameter number.
+      channel (int, optional): Channel to use for communication.
 
     Returns:
       Parameter data if successful, None otherwise.
@@ -159,20 +181,21 @@ class instrument(object):
     try:
       parm = self.db.get_parameter(dde_nr)
     except:
-      raise ValueError('DDE parameter number error!')	  
-    resp = self.read_parameters([parm])	
+      raise ValueError('DDE parameter number error!')
+    resp = self.read_parameters([parm], channel=channel)	
     if resp != None:
       for r in resp:
         return r['data']
     else:
       return None
 
-  def writeParameter(self, dde_nr, data):
+  def writeParameter(self, dde_nr, data, channel=None):
     """Write a single parameter indicated by DDE nr.
 
     Args:
       dde_nr (int): FlowDDE parameter number.
       data: Parameter data to write.
+      channel (int, optional): Channel to use for communication.
 
     Returns:
       True if successful, False otherwise.
@@ -182,33 +205,39 @@ class instrument(object):
     except:
       raise ValueError('DDE parameter number error!')
     parm['data'] = data
-    resp = self.write_parameters([parm])
+    resp = self.write_parameters([parm], channel=channel)
     return (resp == PP_STATUS_OK)
 
-  def read_parameters(self, parameters):
+  def read_parameters(self, parameters, callback=None, channel=None):
     """Read multiple parameters.
     
     Args:
       parameters: List of parameter objects.
+      callback (function, optional): Function to be called when parameters are received (function will return directly!).
+      channel (int, optional): Channel to use for communication.
       
     Returns:
       List with parameters with data if successful, list with one status item otherwise.
     """
     parameters[0]['node'] = self.address
-    return self.master.read_parameters(parameters)
+    parameters = [self.__modify_parameter_channel(parm, channel) for parm in parameters]
+    return self.master.read_parameters(parameters, callback)
 
-  def write_parameters(self, parameters, command=PP_COMMAND_SEND_PARM_WITH_ACK):
+  def write_parameters(self, parameters, command=PP_COMMAND_SEND_PARM_WITH_ACK, callback=None, channel=None):
     """Write multiple parameters.
     
     Args:
       parameters: List of parameter objects, with data.
       command (int, optional): Propar command to use for writing.
+      callback (function, optional): Function to be called when parameters are received (function will return directly!).
+      channel (int, optional): Channel to use for communication.
 
     Returns:
       Propar status code (0 if successful).
     """
     parameters[0]['node'] = self.address
-    return self.master.write_parameters(parameters, command)
+    parameters = [self.__modify_parameter_channel(parm, channel) for parm in parameters]
+    return self.master.write_parameters(parameters, command, callback)
 
   def read(self, proc_nr, parm_nr, parm_type):
     """Read a single parameter.
@@ -247,7 +276,7 @@ class instrument(object):
       True if successful, False otherwise.
     """
     time_char = bytes([0x30+time]).decode('ascii')
-    return self.write(0, 0, PP_TYPE_STRING, time_char)
+    return self.writeParameter(1, time_char)
 
   @property
   def setpoint(self):
@@ -256,7 +285,7 @@ class instrument(object):
     Returns:
       Instrument setpoint if successful, None otherwise.
     """
-    self._setpoint = self.read(1, 1, PP_TYPE_INT16)
+    self._setpoint = self.readParameter(9)
     return self._setpoint
 
   @setpoint.setter
@@ -266,7 +295,7 @@ class instrument(object):
     Returns:
       True if successful, False otherwise.
     """
-    return self.write(1, 1, PP_TYPE_INT16, value)
+    return self.writeParameter(9, value)
 
   @property
   def measure(self):
@@ -275,7 +304,7 @@ class instrument(object):
     Returns:
       Instrument measure if successful, None otherwise.
     """
-    measure = self.read(1, 0, PP_TYPE_BSINT16)
+    measure = self.readParameter(8)
     return measure
 
   @property
@@ -285,7 +314,7 @@ class instrument(object):
     Returns:
       Instrument ID if successful, None otherwise.
     """
-    return self.read(0, 0, PP_TYPE_STRING)
+    return self.readParameter(1)
 
         
 
@@ -407,9 +436,9 @@ class master(object):
       self.response_timeout = org_timeout
 
     while scan_address != 0 and loop_detected == False:
-      parms = [{'node': scan_address, 'proc_nr':   0, 'parm_nr': 1, 'parm_type': PP_TYPE_INT8  },# address of this node
-               {'node': scan_address, 'proc_nr':   0, 'parm_nr': 0, 'parm_type': PP_TYPE_STRING},# id
-               {'node': scan_address, 'proc_nr':   0, 'parm_nr': 3, 'parm_type': PP_TYPE_INT8  }]# address of the next node
+      parms = [{'node': scan_address, 'proc_nr': 0, 'parm_nr':  1, 'parm_type': PP_TYPE_INT8  },# address of this node
+               {'node': scan_address, 'proc_nr': 0, 'parm_nr':  0, 'parm_type': PP_TYPE_STRING},# id
+               {'node': scan_address, 'proc_nr': 0, 'parm_nr':  3, 'parm_type': PP_TYPE_INT8  }]# address of the next node
 
       resp = self.read_parameters(parms)
 
@@ -435,7 +464,7 @@ class master(object):
           print("This is node {:>2} ({:}). Next node is {:}".format(resp[0]['data'], serial_number, resp[2]['data']))
 
         # Try to get device type from device
-        dev_resp = self.read_parameters([{'node': scan_address, 'proc_nr': 113, 'parm_nr': 1, 'parm_type': PP_TYPE_STRING}]) # device type?
+        dev_resp = self.read_parameters([{'node': scan_address, 'proc_nr': 113, 'parm_nr': 1, 'parm_type': PP_TYPE_STRING}]) # device type
         
         if dev_resp[0]['status'] == PP_STATUS_OK:
           device_type = dev_resp[0]['data']
@@ -449,6 +478,14 @@ class master(object):
             if device_type == int(option['value']):
               device_type = option['description'].split(':')[0]
 
+        # Try to get the number of channels from device
+        chan_resp = self.read_parameters([{'node': scan_address, 'proc_nr': 0, 'parm_nr': 18, 'parm_type': PP_TYPE_INT8}]) # number of channels
+        
+        if chan_resp[0]['status'] == PP_STATUS_OK:
+          nr_of_channels = chan_resp[0]['data']
+        else:
+          nr_of_channels = 1
+
         # Scan address = next address
         scan_address = resp[2]['data']
 
@@ -456,11 +493,11 @@ class master(object):
         # In that case there is a network loop, and we stop looking for nodes.
         for node in found_nodes:
           if scan_address == node['address']:
-            loop_detected == True
+            loop_detected = True
             if self.debug:
               print('Found network loop on node {:}'.format(resp[0]['data']))
 
-        found_nodes.append({'address': resp[0]['data'], 'type': device_type, 'serial': serial_number, 'id': resp[1]['data']})        
+        found_nodes.append({'address': resp[0]['data'], 'type': device_type, 'serial': serial_number, 'id': resp[1]['data'], 'channels': nr_of_channels})        
 
     return found_nodes
 
@@ -788,34 +825,28 @@ class database(object):
   """The database class is used to convert FlowDDE numbers to propar parameter objects.
   
     Several other supporting functions are also provided for manual use.
-  
-  Args:
-    database_path (str, optional): Use a custom database json file.
   """
 
   def __init__(self, database_path=None):
     #Columns:
     #Parameter	LongName	Name	Available	Group0	Group1	Group2	Process	FBnr	VarType	VarType2	VarLength	Min	Max	Read	Write	Poll	Advanced	Secured	Highly secured	Default	Description
-    if database_path == None:
-      database_path = os.path.join(os.path.dirname(__file__), "parameters.json")
-    with open(database_path) as f:
-      json_db        = json.load(f)
-      parm_list      = json_db['allparameters']
-      self.parm_vals = json_db['parvalue']
-      self.dde_dict  = {}
-      self.pp_dict   = {}
-      # Create dicts for faster access to parameters
-      for parm in self.__rows_to_parms(parm_list):
-        # Create dde dict
-        self.dde_dict[parm['dde_nr']] = parm
-        # Create propar dict
-        proc_nr = parm['proc_nr']
-        parm_nr = parm['parm_nr']
-        if proc_nr not in self.pp_dict.keys():          
-          self.pp_dict[proc_nr] = {}
-        if parm_nr not in self.pp_dict[proc_nr].keys():
-          self.pp_dict[proc_nr][parm_nr] = []
-        self.pp_dict[proc_nr][parm_nr].append(parm)
+    # Import database
+    parm_list      = parameters.parameters['allparameters']
+    self.parm_vals = parameters.parameters['parvalue']
+    self.dde_dict  = {}
+    self.pp_dict   = {}
+    # Create dicts for faster access to parameters
+    for parm in self.__rows_to_parms(parm_list):
+      # Create dde dict
+      self.dde_dict[parm['dde_nr']] = parm
+      # Create propar dict
+      proc_nr = parm['proc_nr']
+      parm_nr = parm['parm_nr']
+      if proc_nr not in self.pp_dict.keys():          
+        self.pp_dict[proc_nr] = {}
+      if parm_nr not in self.pp_dict[proc_nr].keys():
+        self.pp_dict[proc_nr][parm_nr] = []
+      self.pp_dict[proc_nr][parm_nr].append(parm)
 
 
   def __rows_to_parms(self, rows):
@@ -1096,24 +1127,29 @@ class _propar_builder(object):
             if (max_message_len - pos) >= 1:
               len_pos = pos
               pos += 1
+              # get bytes
+              if type(parameter['data']) is str:
+                str_bytes = parameter['data'].encode('utf-8')
+              elif type(parameter['data']) is bytes:
+                str_bytes = parameter['data']
+              else:
+                str_bytes = str(parameter['data']).encode('utf-8')
               # get string length
               if parameter['parm_size'] == 0:
                 try:
-                  len_str = len(parameter['data'])
+                  len_str = len(str_bytes)
                 except:
                   pass
               else:
                 len_str = parameter['parm_size']
               message[len_pos] = len_str
+              # pad with spaces if needed
+              if len_str > len(str_bytes):
+                str_bytes += b' ' * (len_str - len(str_bytes))
               # adjust string length to parm_size
-              for char in parameter['data'].ljust(len_str):
-                if type(char) is str: # is still a string convert to int of byte
-                  char = int.from_bytes(bytes(char, encoding='ascii'), byteorder='little')
-                message[pos] = char
+              for c in str_bytes:
+                message[pos] = c
                 pos += 1
-                len_str -= 1
-                if len_str == 0:
-                  break
               # zero terminate the string
               message[pos] = 0
               pos += 1
@@ -1331,7 +1367,7 @@ class _propar_builder(object):
               # Decode string and store data
               string_bytes = bytes(message[pos:pos+parameter['parm_size']])
               try:
-                parameter['data'] = string_bytes.decode('ascii')
+                parameter['data'] = string_bytes.decode('utf-8')
               except:
                 parameter['data'] = string_bytes
 
@@ -1458,7 +1494,12 @@ class _propar_provider(object):
     # External flags / options
     self.debug = debug
     self.dump  = dump
-    self.mode  = mode
+    self.mode  = mode    
+	
+	# Additional features
+    self.auto_reopen = True
+    self.open_count  = 0
+    self.dump_byte   = None # function that takes dumped byte / char.
 
     # queues for propar data packets
     self.__receive_queue  = collections.deque()
@@ -1512,8 +1553,6 @@ class _propar_provider(object):
     This function must run in a thread.
     """
     while self.run:
-      # try-except added to fix issues when we are stopped (comport is closed).
-  	  # due to thread, this can cause read to error out.	  
       try:
         if self.paused == False:
           if self.serial.in_waiting:
@@ -1522,13 +1561,25 @@ class _propar_provider(object):
               was_propar_byte = self.__process_propar_byte(data_byte)
               if self.dump != 0:
                 if self.dump == 2 or was_propar_byte == False:
-                  print(chr(data_byte), end='')
+                  if self.dump_byte:
+                    self.dump_byte(chr(data_byte))
+                  else:
+                    print(chr(data_byte), end='')
           else:
             time.sleep(0.001)
           if self.dump != 0:
             print(end='', flush=True)
         else:
           time.sleep(0.002)
+      except (IOError, TypeError) as e:
+        time.sleep(0.2)
+        if self.auto_reopen:
+          try:
+            self.serial.close()
+            self.serial.open()
+            self.open_count += 1
+          except:
+            pass
       except:
         time.sleep(0.002)
 
